/** 3DGPL *************************************************\
 * ()                                                     *
 * Functions for indoor volume objects.                   *
 *                                                        *
 * Ifdefs:                                                *
 *  _CI_                    Colour/Intensity model;       *
 *  _RGB_                   RGB model;                    *
 *  _Z_BUFFER_              Depth array;                  *
 *  _PAINTER_               Back-front order.             *
 *                                                        *
 * Defines:                                               *
 *  M_init_volume_object    Compute normal vectors;       *
 *  M_shade_volume_object   Flat and interpolative;       *
 *  M_render_volume_object  Renders a polygonal solid.    *
 *                                                        *
 * Internals:                                             *
 *  MI_init_volumes         Recursively all volumes;      *
 *  MI_shade_volumes        Recursively all volumes;      *
 *  MI_render_volumes       Recursively all volumes;      *
 *                                                        *
 * (c) 1995-98 Sergei Savchenko, (savs@cs.mcgill.ca)      *
\**********************************************************/

#include "../graphics/graphics.h"           /* 2-D rendering */
#include "../trans/trans.h"                 /* 3-D transformations */
#include "../engine/engine.h"               /* 3-D engine */
#include "../clipper/clipper.h"             /* get/set_bounds */
#include <stdlib.h>                         /* NULL */
#include <limits.h>                         /* INT_MAX etc */

extern int M_no_vertices;                   /* borrowed polygonal stuff */
extern int *M_vertices;                     /* static vertices X Y Z */
int MI_add_tuple(int *vertex);              /* eng-poly.c */

int M_no_volumes;                           /* stack of volumes being drawn */
struct M_volume_object *M_volumes[M_MAX_VOLUMES];

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Building normals for all volumes.                     *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void MI_init_volumes(struct M_volume_object *object)
{
 int tmp_normals[M_MAX_OBJECT_POLYGONS*T_LNG_VECTOR];
 int normals[M_MAX_OBJECT_POLYGONS*T_LNG_VECTOR];
 int i,j,k,n[5],num,norm_idx;

 for(i=0;i<M_no_volumes;i++)
  if(M_volumes[i]==object) return;          /* this one is already inited */

 M_volumes[M_no_volumes++]=object;          /* remember this one's inited */
 if(M_no_volumes>=M_MAX_VOLUMES)
  HW_error("(Engine) Can't handle this many volumes.");

 M_vertices=tmp_normals; M_no_vertices=0;   /* nothing added yet */

 for(i=0;i<object->m_no_polygons;i++)       /* for all polygons */
 {
  T_normal_plane(object->m_vertices+object->m_polygons[i]->m_vertices[0]*T_LNG_VECTOR,
                 object->m_vertices+object->m_polygons[i]->m_vertices[M_LNG_POLYGON_VERTEX]*T_LNG_VECTOR,
                 object->m_vertices+object->m_polygons[i]->m_vertices[M_LNG_POLYGON_VERTEX*2]*T_LNG_VECTOR,
                 normals+i*T_LNG_VECTOR
                );
  if(object->m_polygons[i]->m_type&M_PLANNAR)
  {                                         /* same normal in flat polygons */
   object->m_polygons[i]->m_normals=(int*)malloc(sizeof(int)*(object->m_polygons[i]->m_no_edges+1));
   if(object->m_polygons[i]->m_normals==NULL)
    HW_error("(Engine) Not enough memory.\n");
   object->m_polygons[i]->m_normals[0]=MI_add_tuple(normals+i*T_LNG_VECTOR);
   for(j=1;j<=object->m_polygons[i]->m_no_edges;j++)
    object->m_polygons[i]->m_normals[j]=object->m_polygons[i]->m_normals[0];
  }
 }

 for(i=0;i<object->m_no_vertices;i++)       /* for all vertices */
 {
  n[0]=n[1]=n[2]=0;
  for(num=0,j=0;j<object->m_no_polygons;j++)/* check all polygons which */
  {
   if(object->m_polygons[j]->m_type&M_CURVED)
   {                                        /* emulate curved surfaces */
    if(object->m_polygons[j]->m_normals==NULL)
    {
     object->m_polygons[j]->m_normals=(int*)malloc(sizeof(int)*(object->m_polygons[j]->m_no_edges+1));
     if(object->m_polygons[j]->m_normals==NULL)
      HW_error("(Engine) Not enough memory.\n");
     HW_set_int(object->m_polygons[j]->m_normals,object->m_polygons[j]->m_no_edges+1,0);
    }

    for(k=0;k<=object->m_polygons[j]->m_no_edges;k++)
    {
     if(object->m_polygons[j]->m_vertices[k*M_LNG_POLYGON_VERTEX]==i)
     {                                      /* if the vertex belong to it */
      object->m_polygons[j]->m_normals[k]=-1;
      n[0]+=normals[j*T_LNG_VECTOR];        /* mark the normal */
      n[1]+=normals[j*T_LNG_VECTOR+1];      /* polygon's normal */
      n[2]+=normals[j*T_LNG_VECTOR+2];
      num++;
     }
    }
   }
  }
  if(num!=0)                                /* for curved only */
  {
   n[0]/=num; n[1]/=num; n[2]/=num;
   norm_idx=MI_add_tuple(n);

   for(j=0;j<object->m_no_polygons;j++)
   {
    if(object->m_polygons[j]->m_type&M_CURVED)
    {                                        /* emulate curved surfaces */
     for(k=0;k<=object->m_polygons[j]->m_no_edges;k++)
     {
      if(object->m_polygons[j]->m_normals[k]==-1)
       object->m_polygons[j]->m_normals[k]=norm_idx;
     }
    }
   }
  }
 }
 object->m_normals=(int*)malloc(sizeof(int)*M_no_vertices*T_LNG_VECTOR);
 object->m_no_normals=M_no_vertices;        /* store the normals */
 HW_copy_int(M_vertices,object->m_normals,M_no_vertices*T_LNG_VECTOR);

 for(i=0;i<object->m_no_gates;i++)          /* recurse */
  MI_init_volumes(object->m_gates[i]->m_volume);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Shading all interconnected volumes.                   *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void MI_shade_volumes(struct M_volume_object *object,
                      int no_lights,
                      struct M_light **lights
                     )
{
 int i;

 if((object->m_no_vertices>=M_MAX_OBJECT_VERTICES)||
    (object->m_no_normals>=M_MAX_OBJECT_VERTICES)
   )
  HW_error("(Engine) Can't handle this many vertices or normals in polygons.");

 for(i=0;i<M_no_volumes;i++)
  if(M_volumes[i]==object) return;          /* this one is already shaded */

 M_volumes[M_no_volumes++]=object;          /* remember this one's shaded */
 if(M_no_volumes>=M_MAX_VOLUMES)
  HW_error("(Engine) Can't handle this many volumes.");

 for(i=0;i<object->m_no_polygons;i++)
 {
  M_shade_polygon(object->m_polygons[i],object->m_vertices,
                  object->m_normals,no_lights,lights
                 );
 }

 for(i=0;i<object->m_no_gates;i++)          /* recurse */
  MI_shade_volumes(object->m_gates[i]->m_volume,no_lights,lights);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Rendering an object connected through this gate.      *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void MI_render_volumes(struct M_volume_object *object,int level)
{
 int tmp_vertices[M_MAX_VOLUME_VERTICES*T_LNG_VECTOR];
 int i;
 int minx,miny,maxx,maxy;                   /* old clipping bounds */

 if(object->m_no_vertices>=M_MAX_VOLUME_VERTICES)
  HW_error("(Engine) Can't handle this many vertices.");

 if(level==0) return;                       /* rendering deapth reached */
 for(i=0;i<M_no_volumes;i++)
  if(M_volumes[i]==object) return;          /* this one is already rendered */

 M_volumes[M_no_volumes++]=object;          /* remember this one's rendered */
 if(M_no_volumes>=M_MAX_VOLUMES)
  HW_error("(Engine) Can't handle this many volumes.");

 T_translation(object->m_vertices,tmp_vertices,object->m_no_vertices,
               M_camera_x,M_camera_y,M_camera_z
              );
 T_world_rotation(tmp_vertices,tmp_vertices,object->m_no_vertices);

 for(i=0;i<object->m_no_gates;i++)          /* render through all gates */
 {
  C_get_bounds(&minx,&miny,&maxx,&maxy);    /* current clipping */

  if(M_set_gate(object->m_gates[i],tmp_vertices))
   MI_render_volumes(object->m_gates[i]->m_volume,level-1);

  C_init_clipping(minx,miny,maxx,maxy);     /* restore original bounds */
 }

 for(i=0;i<object->m_no_polygons;i++)
 {
  M_render_polygon(object->m_polygons[i],tmp_vertices,NULL,NULL);
 }
}

/**********************************************************\
 * Initialising all volumes.                              *
\**********************************************************/

void M_init_volume_object(struct M_volume_object *object)
{
 M_no_volumes=0;
 MI_init_volumes(object);
}

/**********************************************************\
 * Shading all volumes.                                   *
\**********************************************************/

void M_shade_volume_object(struct M_volume_object *object,
                           int no_lights,
                           struct M_light **lights
                          )
{
 M_no_volumes=0;
 MI_shade_volumes(object,no_lights,lights);
}

/**********************************************************\
 * Rendering a volume object in perspective.              *
 * NB! beware that it is assumed that any two volumes are *
 * connected through a single portal only.                *
\**********************************************************/

void M_render_volume_object(struct M_volume_object *object,int depth)
{
 M_no_volumes=0;                            /* nothing rendered yet */
 MI_render_volumes((struct M_volume_object*)object,depth);
}

/**********************************************************/
